
int_21h:				; AH = function number
	push	ds es
	pusha
	cli
	cld
	push	FLAT_SEL
	pop	ds
	push	ds
	pop	es
	or	ah,ah			; function 00h:
	jz	get_int_vector		;  get interrupt vector
	cmp	ah,01h			; function 01h:
	je	set_int_vector		;  set interrupt vector
	cmp	ah,10h			; function 10h:
	je	allocate_memory 	;  allocate memory block
	cmp	ah,11h			; function 11h:
	je	free_memory		;  free memory block
	cmp	ah,12h			; function 12h:
	je	resize_memory		;  resize memory block
	cmp	ah,20h			; function 20h:
	je	create_descriptor	;  create descriptor
	cmp	ah,21h			; function 21h:
	je	free_descriptor 	;  free descriptor
	cmp	ah,30h			; function 30h:
	je	reboot_computer 	;  reboot computer
	cmp	ah,31h			; function 31h:
	je	restart_system		;  restart_system
	mov	[error_code],_function_unavailable
	jmp	int_error
  get_int_vector:			; AL = interrupt number
	movzx	eax,al
	shl	eax,3
	add	eax,10800h		; add address of IDT
	cli
	mov	cx,word [ds:eax+2]	; get selector
	mov	[esp+18h],cx		; return selector in CX
	mov	dx,word [ds:eax+6]
	shl	edx,16
	mov	dx,word [ds:eax]
	mov	[esp+14h],edx		; return offset in EDX
	sti
	jmp	int_ok
  set_int_vector:			; CX:EDX - interrupt handler
	movzx	eax,al			; AL = interrupt number
	shl	eax,3
	add	eax,10800h		; add address of IDT
	cli
	mov	word [ds:eax+2],cx	; store selector
	mov	word [ds:eax],dx	; store offset
	shr	edx,16
	mov	word [ds:eax+6],dx
	sti
	jmp	int_ok
  allocate_memory:			; ECX = size
	xor	ebp,ebp
	mov	[error_code],_invalid_value
	or	ecx,ecx 		; size is 0?
	jz	int_error		; if yes, return error
	dec	ecx			; convert size to pages count
	shr	ecx,12
	inc	ecx
	mov	esi,11000h		; address of first page table entry
	call	get_pagetable		; get first page table
      search_linear_space:
	xor	ebx,ebx
	lea	eax,[esi-4]
	mov	edx,edi
      scan_linear_page:
	cmp	edi,1F000h+1000h	; end of page table?
	je	next_pagetable		; if yes, get next
	add	edi,4
	cmp	dword [edi-4],0 	; is this page unused?
	jne	search_linear_space	; if no, search for next unused linear block
	inc	ebx			; increase pages count
	cmp	ebx,ecx 		; got needed pages count?
	jne	scan_linear_page	; if no, check for next unused page
	jmp	allocate_physical	; allocate physical memory
      next_pagetable:
	cmp	esi,5FFCh		; end of page directory?
	je	allocate_physical	; if yes, allocate physical memory
	call	get_pagetable		; get next page table
	jnc	scan_linear_page	; if ok, check for next unused page
      allocate_physical:
	mov	esi,eax 		; get page table which contains first
	call	get_pagetable		;  page of allocated linear block
	mov	edi,edx
	mov	edx,eax 		; calculate address of allocated
	sub	edx,11000h		;  linear block
	shl	edx,20
	mov	eax,edi
	sub	eax,1F000h
	shl	eax,10
	or	edx,eax
	mov	[esp+14h],edx		; return address in EDX
	mov	ebx,20000h		; get address of memory table
      search_physical_space:
	cmp	ebx,[extensions_memory] ; end of memory table?
	je	mem_allocated		; if yes, finish allocating
	inc	ebx
	cmp	byte [ebx-1],0		; is this page free?
	jne	search_physical_space	; if no, check next page
	mov	byte [ebx-1],3		; mark page as used
	mov	eax,ebx 		; calculate address of this page
	sub	eax,20000h + 1
	shl	eax,12
      make_page_entry:
	cmp	edi,1F000h+1000h	; end of current page table?
	je	make_in_next_pagetable	; if yes, create entry in next table
	or	eax,1 shl 9 + 111b	; create page entry
	or	ebp,ebp 		; first page in block?
	jnz	store_page_entry	; if no, store this entry
	bts	eax,10			; mark entry as first page
	not	ebp
      store_page_entry:
	bt	dword [edi],9		; is this page marked as allocated?
	jc	allocation_broken	; if yes, stop allocating
	stosd				; store page entry in current table
	loop	search_physical_space	; search for next free page, if needed
	jmp	mem_allocated
      make_in_next_pagetable:
	cmp	esi,5FFCh		; end of page directory?
	je	mem_allocated		; if yes, finish allocating
	call	get_pagetable		; get next page table
	jnc	make_page_entry 	; if ok, add this page to linear block
	or	eax,1b			; use this page as new page table
	mov	[esi],eax
	mov	byte [ebx-1],1		; mark page as system
	mov	eax,cr3
	mov	cr3,eax 		; reload CR3 to clear TLB
	call	get_pagetable
	push	ecx edi
	xor	eax,eax 		; clear new page table
	mov	ecx,1024
	rep	stosd
	pop	edi ecx
	jmp	search_physical_space	; search for next free page
      allocation_broken:
	mov	byte [ebx-1],0		; mark page as free
	bts	dword [edi],11		; mark page as last in block
	sub	edi,4
	jnc	allocation_finished
	sub	esi,4
	add	edi,1000h
	jmp	allocation_finished
      mem_allocated:
	bts	dword [edi-4],11	; mark page as last in block
      allocation_finished:
	lea	ecx,[esi-4]		; calculate linear address of first
	sub	ecx,11000h		;  page after allocated block
	shl	ecx,20
	mov	eax,edi
	sub	eax,1F000h
	shl	eax,10
	or	ecx,eax
	sub	ecx,edx 		; calculate size of allocated block
	mov	[esp+18h],ecx		; return size in ECX
	jmp	int_ok
      get_pagetable:
	push	eax
	lodsd				; get page table entry
	or	eax,eax 		; does page table exist?
	jz	no_pagetable		; if no, return error
	mov	edi,12000h + 1Fh*4	; map page table in window
	stosd				;  at address 1F000h
	mov	eax,cr3
	mov	cr3,eax 		; reload CR3 to clear TLB
	mov	edi,1F000h
	pop	eax
	clc
	ret
      no_pagetable:
	sub	esi,4
	pop	eax
	stc
	ret
  free_memory:				; EDX = linear address
	mov	[error_code],_invalid_address
	test	edx,(1 shl 12)-1	; check for valid address
	jnz	int_error
	mov	esi,edx 		; get page table which contains
	shr	esi,20			;  first page of block to free
	add	esi,11000h
	call	get_pagetable
	mov	eax,edx 		; get first page to free
	shr	eax,10
	and	eax,(1024-1) shl 2
	add	edi,eax
	bt	dword [edi],10		; is this page marked as first?
	jnc	int_error		; if no, error
      free_page:
	cmp	edi,1F000h+1000h	; end of page table?
	jne	remove_page_entry
	call	get_pagetable		; if yes, get next page table
	jc	int_error		; if failed, return error
      remove_page_entry:
	bt	dword [edi],9		; is this page marked as allocated?
	jnc	int_error		; if no, return error
	mov	eax,[edi]		; get physical address of page
	shr	eax,12
	add	eax,20000h
	cmp	byte [eax],3		; is this page marked as used?
	jne	int_error		; if no, return error
	mov	byte [eax],0		; mark page as free
	mov	edx,[edi]
	xor	eax,eax 		; clear page entry
	stosd
	bt	edx,11			; was it last page in block?
	jnc	free_page		; if no, free next page
	jmp	int_ok
  resize_memory:			; EDX = linear address, ECX = new size
	or	ecx,ecx 		; new size is 0?
	jz	free_memory		; if yes, free block
	mov	[error_code],_invalid_address
	test	edx,(1 shl 12)-1	; check for valid address
	jnz	int_error
	dec	ecx			; convert size to pages count
	shr	ecx,12
	inc	ecx
	mov	ebp,ecx
	mov	esi,edx 		; get page table which contains
	shr	esi,20			;  first page of block
	add	esi,11000h
	call	get_pagetable
	mov	eax,edx 		; get first page
	shr	eax,10
	and	eax,(1024-1) shl 2
	add	edi,eax
	bt	dword [edi],10		; is this page marked as first?
	jnc	int_error		; if no, error
      resize_block:
	cmp	edi,1F000h+1000h	; end of page table?
	jne	check_page
	call	get_pagetable		; if yes, get next page table
	jc	int_error		; if failed, return error
      check_page:
	bt	dword [edi],11		; is this page marked as last?
	jc	add_pages		; if yes, allocate new pages
	add	edi,4
	loop	resize_block
	bts	dword [edi-4],11	; mark page as last
	shl	ebp,12
	mov	[esp+18h],ebp		; return new size in ECX
	jmp	free_page		; free rest of pages
      add_pages:
	dec	ecx
	or	ecx,ecx 		; no size change?
	jz	int_ok			; if yes, resizing done
	and	dword [edi],not (1 shl 11)
	add	edi,4
	mov	ebx,20000h		; get address of memory table
	jmp	search_physical_space
  get_page:				; get page for system use
	mov	esi,20000h		; get address of memory table
      find_free_page:
	lodsb
	or	al,al			; found free page?
	jnz	find_free_page		; if no, check next
	mov	byte [esi-1],1		; mark page as system
	sub	esi,20000h + 1		; create page entry
	shl	esi,12
	mov	eax,esi
	or	eax,111b
	ret
  create_descriptor:			; EBX = base, EDX = limit
	mov	esi,10000h		; CX = access rights
	mov	[error_code],_out_of_resources
      find_free_descriptor:
	add	esi,8
	cmp	esi,10800h
	je	int_error
	test	byte [esi+6],10000b	; descriptor marked as used?
	jnz	find_free_descriptor	; if no, check next
	mov	[esi+2],ebx		; store segment base
	shr	ebx,24
	mov	[esi+7],bl		; store high bits of segment base
	and	ch,1 shl 6
	or	cl,10010000b
	or	ch,10000b		; mark descriptor as used
	mov	[esi+5],cx		; store access rights word
	cmp	edx,1 shl 20		; big segment?
	jb	set_segment_limit	; if no, store current limit
	shr	edx,12
	or	byte [esi+6],1 shl 7	; set granularity bit
      set_segment_limit:
	mov	[esi],dx		; store low bits of segment limit
	shr	edx,16
	or	[esi+6],dl		; store high bits of segment limit
	sub	esi,10000h
	mov	[esp+18h],si		; return selector in CX
	jmp	int_ok
  free_descriptor:			; CX = selector
	mov	[error_code],_invalid_selector
	test	cx,100b 		; descriptor in GDT?
	jnz	int_error		; if no, return error
	and	cx,not 111b
	movzx	ecx,cx
	add	ecx,10000h		; get address of descriptor
	test	byte [ecx+6],10000b	; descriptor marked as used?
	jz	int_error		; if no, return error
	xor	eax,eax 		; clear descriptor data
	mov	[ecx],eax
	mov	[ecx+4],eax
	jmp	int_ok
  reboot_computer:
	mov	al,0FEh
	out	64h,al			; reboot computer
  restart_system:
	mov	eax,cr0
	and	eax,not (1 shl 31)
	mov	cr0,eax 		; disable paging
;        mov     cl,[cs:kernel_size]     ; get kernel size
;        jmp     kernel_init             ; restart kernel
