
init_video_memory:
	call	get_page
	mov	[video_page],eax
	mov	edi,12000h + 1Fh*4	; map this page in window
	stosd				;  at address 1F000h
	mov	edi,1F000h
	mov	[11FF0h],eax		; create page table entry
	mov	ecx,1024
	xor	eax,eax 		; clear page entries
	rep	stosd
	mov	eax,cr3
	mov	cr3,eax 		; reload CR3 to clear TLB
	ret

int_23h:
	push	ds es
	pusha
	cli
	cld
	push	FLAT_SEL
	pop	ds
	push	ds
	pop	es
	or	ah,ah			; function 00h:
	jz	set_text_mode		;  set text mode
	cmp	ah,01h			; function 01h:
	je	set_8bit_mode		;  set 8-bit graphic mode
	cmp	ah,02h			; function 02h:
	je	unsupported_mode	;  set 16-bit graphic mode
	cmp	ah,03h			; function 03h:
	je	unsupported_mode	;  set 24-bit graphic mode
	cmp	ah,04h			; function 04h:
	je	unsupported_mode	;  set 32-bit graphic mode
	cmp	ah,10h			; function 10h:
	je	get_video_mode		;  get video mode
	cmp	ah,20h			; function 20h:
	je	display_character	;  display character
	cmp	ah,21h			; function 21h:
	je	display_string		;  display string
	cmp	ah,22h			; function 22h:
	je	get_cursor_position	;  get cursor position
	cmp	ah,23h			; function 23h:
	je	set_cursor_position	;  set cursor position
	cmp	ah,24h			; function 22h:
	je	get_cursor_shape	;  get cursor shape
	cmp	ah,25h			; function 23h:
	je	set_cursor_shape	;  set cursor shape
	cmp	ah,80h			; function 80h:
	je	set_video_memory	;  map video memory
	mov	[error_code],_function_unavailable
	jmp	int_error
  set_text_mode:
	cmp	dx,1950h
	je	set_80x25_text_mode
      unsupported_mode:
	mov	[error_code],_invalid_value
	jmp	int_error
  set_8bit_mode:
	jmp	unsupported_mode
  set_video_memory:
	mov	[error_code],_invalid_value
	cmp	ecx,400000h		; exceeded maximum size?
	ja	int_error		; if yes, return error
	call	map_video_memory
	jmp	int_ok
  map_video_memory:			; EDX = physical address, ECX = size
	dec	ecx
	shr	ecx,12
	inc	ecx
	mov	eax,[video_page]
	mov	edi,12000h + 1Fh*4	; map video page table in window
	stosd				;  at address 1F000h
	mov	eax,cr3
	mov	cr3,eax 		; reload CR3 to clear TLB
	mov	edi,1F000h
	mov	eax,edx
	mov	edx,1024
	sub	edx,ecx
      map_video_page:			; create page entries
	or	eax,111b
	stosd
	add	eax,1000h
	loop	map_video_page
	mov	ecx,edx
	xor	eax,eax
	rep	stosd			; clear rest of page entries
	ret
  get_video_mode:
	mov	al,[mode_type]
	mov	[esp+1Ch+1],al		  ; return mode type in AH
	or	al,al
	jz	get_text_resolution
	mov	ax,[horizontal_resolution]
	mov	[esp+14h],ax		; return horizontal resolution in DX
	mov	ax,[vertical_resolution]
	mov	[esp+18h],ax		; return vertical resolution in CX
	jmp	int_ok
      get_text_resolution:
	mov	al,byte [horizontal_resolution]
	mov	[esp+14h],al		; return horizontal resolution in DL
	mov	al,byte [vertical_resolution]
	mov	[esp+14h+1],al		; return vertical resolution in DH
	jmp	int_ok
  display_string:
	mov	[error_code],_function_unavailable
	cmp	[mode_type],0		; current video mode is text?
	jne	int_error		; if no, return error
	mov	ds,[esp+24h]		; get caller's DS
	mov	esi,edx
      draw_characters:
	lodsb
	or	al,al
	jz	int_ok
	mov	ah,[esp+1Ch]		; get color from caller's AL
	push	ds esi
	push	FLAT_SEL
	pop	ds
	call	draw_character
	pop	esi ds
	jmp	draw_characters
  display_character:			; DL = character, AL = color
	mov	[error_code],_function_unavailable
	cmp	[mode_type],0		; current video mode is text?
	jne	int_error		; if no, return error
	mov	ah,al
	mov	al,dl
	call	draw_character
	jmp	int_ok
      draw_character:			; AX = character and color
	push	eax
	call	get_cpos
	pop	eax
	cmp	al,0Dh
	je	carriage_return
	cmp	al,0Ah
	je	line_feed
	cmp	al,8
	je	backspace
	movzx	edi,cx
	shl	edi,1
	cmp	edi,[video_memory_size]
	jae	position_over_screen
	add	edi,0FF000000h
	stosw
	inc	ecx
	movzx	eax,[horizontal_resolution]
	movzx	ebx,[vertical_resolution]
	imul	eax,ebx
	cmp	cx,ax
	je	add_text_line
	call	set_cpos
	ret
      position_over_screen:
	add	esp,4
	mov	[error_code],_invalid_value
	jmp	int_error
      carriage_return:
	movzx	eax,cx
	xor	edx,edx
	movzx	ebx,[horizontal_resolution]
	div	ebx
	mul	ebx
	mov	ecx,eax
	call	set_cpos
	ret
      line_feed:
	movzx	eax,[horizontal_resolution]
	add	cx,ax
	movzx	ebx,[vertical_resolution]
	imul	eax,ebx
	cmp	cx,ax
	je	add_text_line
	call	set_cpos
	ret
      backspace:
	movzx	eax,cx
	xor	edx,edx
	movzx	ebx,[horizontal_resolution]
	div	ebx
	or	edx,edx
	jz	backspace_ok
	dec	cx
	call	set_cpos
      backspace_ok:
	ret
      add_text_line:
	call	wait_for_retrace
	sub	cx,[horizontal_resolution]
	call	set_cpos
      scroll_text:
	mov	esi,0FF000000h
	mov	edi,esi
	movzx	ebx,[horizontal_resolution]
	lea	esi,[esi+ebx*2]
	movzx	ecx,[vertical_resolution]
	dec	ecx
	imul	ecx,ebx
	rep	movsw
	mov	ecx,ebx
	mov	ax,0720h
	rep	stosw
	ret
  get_cursor_position:
	mov	[error_code],_function_unavailable
	cmp	[mode_type],0		; current video mode is text?
	jne	int_error		; if no, return error
	call	get_cpos
	mov	eax,ecx
	xor	edx,edx
	movzx	ebx,[horizontal_resolution]
	div	ebx			; DL = horizontal position
	mov	dh,al			; DH = vertical position
	mov	[esp+14h],dx		; return DX
	jmp	int_ok
      get_cpos:
	mov	dx,3D4h
	mov	al,0Eh
	out	dx,al
	inc	dx			; DX = 3D5h
	in	al,dx
	mov	ch,al
	dec	dx			; DX = 3D4h
	mov	al,0Fh
	out	dx,al
	inc	dx			; DX = 3D5h
	in	al,dx
	mov	cl,al			; CX = current cursor position
	ret
  set_cursor_position:
	mov	[error_code],_function_unavailable
	cmp	[mode_type],0		; current video mode is text?
	jne	int_error		; if no, return error
	movzx	eax,dh			; DH = new vertical position
	movzx	ecx,dl			; DL = new horizontal position
	movzx	ebx,[horizontal_resolution]
	mul	ebx
	add	ecx,eax
	call	set_cpos
	jmp	int_ok
      set_cpos: 			; CX = new cursor position
	mov	dx,3D4h
	mov	al,0Eh
	out	dx,al
	inc	dx			; DX = 3D5h
	mov	al,ch
	out	dx,al
	dec	dx			; DX = 3D4h
	mov	al,0Fh
	out	dx,al
	inc	dx			; DX = 3D5h
	mov	al,cl
	out	dx,al
	ret
  get_cursor_shape:
	mov	[error_code],_function_unavailable
	cmp	[mode_type],0		; current video mode is text?
	jne	int_error		; if no, return error
	mov	dx,3D4h
	mov	al,0Ah
	out	dx,al
	inc	dx			; DX = 3D5h
	in	al,dx
	mov	[esp+18h+1],al		; return cursor start in CH
	dec	dx			; DX = 3D4h
	mov	al,0Bh
	out	dx,al
	inc	dx			; DX = 3D5h
	in	al,dx
	mov	[esp+18h],al		; return cursor end in CL
	jmp	int_ok
  set_cursor_shape:			; CH = cursor start, CL = cursor end
	mov	[error_code],_function_unavailable
	cmp	[mode_type],0		; current video mode is text?
	jne	int_error		; if no, return error
	call	set_cshape
	jmp	int_ok
      set_cshape:
	mov	dx,3D4h
	mov	al,0Ah
	out	dx,al
	inc	dx			; DX = 3D5h
	mov	al,ch
	out	dx,al
	dec	dx			; DX = 3D4h
	mov	al,0Bh
	out	dx,al
	inc	dx			; DX = 3D5h
	mov	al,cl
	out	dx,al
	ret
  wait_for_retrace:
	mov	dx,3DAh
	in	al,dx
	test	al,1000b
	je	wait_for_retrace
	ret

  set_80x25_text_mode:			; [temporary procedure]
	mov	edx,0B8000h
	mov	ecx,80*25*2
	mov	[video_memory_size],ecx
	call	map_video_memory
	mov	[mode_type],0		; text mode
	mov	[horizontal_resolution],80
	mov	[vertical_resolution],25
	mov	edi,0FF000000h
	mov	ax,0720h
	mov	ecx,80*25
	rep	stosw
	mov	cx,0E0Fh
	call	set_cshape
	xor	cx,cx
	call	set_cpos
	jmp	int_ok

video_page dd 0
video_memory_size dd 0
mode_type db 0
horizontal_resolution dw 0
vertical_resolution dw 0
